
#include "Threading.h"

#ifndef WIN32

bool cond_wait_timed(condition_t * c, int ticks)
{
    struct timeval now;
    struct timespec timeout;
    int rv;
	
    pthread_mutex_lock(&c->mtx);
    gettimeofday(&now, NULL);
    timeout.tv_sec = now.tv_sec + ((ticks) / 1000);
    timeout.tv_nsec = now.tv_usec * 1000 + ((ticks) % 1000) * 1000000;
    if (timeout.tv_nsec > 1000000000) {
        ++timeout.tv_sec;
        timeout.tv_nsec -= 1000000000;
    }
    rv = pthread_cond_timedwait(&c->cond, &c->mtx, &timeout);
    pthread_mutex_unlock(&c->mtx);
  
    if (rv == 0)
		return true;
	return false;
}

 
void sleep_ms( long msec )
{
#ifdef HAVE_NANOSLEEP
    struct timespec req = {0}, rem = {0};
    PaTime time = msec / 1.e3;
    req.tv_sec = (time_t)time;
	// This was an assert, but just give up the sleep attemp instead
    if (time - req.tv_sec >= 1.0)
		return;
    req.tv_nsec = (long)((time - req.tv_sec) * 1.e9);
    nanosleep(&req, &rem);
    // XXX: Try sleeping the remaining time (contained in rem) if interrupted by a signal? 
#else
    while( msec > 999 )     // For OpenBSD and IRIX, argument 
        {                   // to usleep must be < 1000000.   
        usleep( 999000 );
        msec -= 999;
        }
    usleep( msec * 1000 );
#endif
}

void set_thread_priority(mythread_t t, int iPriority)
{
	struct sched_param 	sParam;
	int 				iPolicy 		= 0;
	int					iNewPriority	= 0;
	const int			PRIORITY_BUMP 	= 2;

	memset(&sParam, 0, sizeof(sParam));
	pthread_getschedparam(t, &iPolicy, &sParam);

	// We bump the priority a little above or below the current level
	iNewPriority = sParam.sched_priority;
	switch (iPriority)
	{
		case MYTHREAD_PRIORITY_BELOW_NORMAL:
		    iNewPriority -= PRIORITY_BUMP;
			if (iNewPriority < 	sched_get_priority_min(SCHED_OTHER))
				iNewPriority = sched_get_priority_min(SCHED_OTHER);
			sParam.sched_priority = iNewPriority;
			pthread_setschedparam(t, SCHED_OTHER, &sParam);
			break;
		case MYTHREAD_PRIORITY_NORMAL:
			break;
		case MYTHREAD_PRIORITY_ABOVE_NORMAL:
		    iNewPriority += PRIORITY_BUMP;
			if (iNewPriority < 	sched_get_priority_max(SCHED_OTHER))
				iNewPriority = sched_get_priority_max(SCHED_OTHER);
			sParam.sched_priority = iNewPriority;
			pthread_setschedparam(t, SCHED_OTHER, &sParam);
			break;
	};
}

// Timed mutex lock, returns true if acquired.
bool mutex_lock_timed(mutex_t* pMutex, int iTimeoutMs)
{
	for (unsigned int i = 0; i < iTimeoutMs; i += 25)
	{
		if (pthread_mutex_trylock(pMutex) == 0)
			return true;
		sleep_ms(25);
	}
	return false;
}

// Create a thread with the given amount of stack size
void create_thread_stack(pthread_t* tt, void *(proc)(void*), int size)
{
	pthread_attr_t attr;
	pthread_attr_init(&attr);  
	pthread_attr_setstacksize(&attr, (size_t) size);
	pthread_create(tt, &attr, proc, NULL);
}

// Create a thread with the given stack size and startup argument
void create_thread_stack_arg(pthread_t* tt, void *(proc)(void*), int size, void* arg)
{
	pthread_attr_t attr;
	pthread_attr_init(&attr);  
	pthread_attr_setstacksize(&attr, (size_t) size);
	pthread_create(tt, &attr, proc, arg);
}

#else

// Win32 implementations
void set_thread_priority(mythread_t t, int iPriority)
{
	switch (iPriority)
	{
		case MYTHREAD_PRIORITY_BELOW_NORMAL:
			SetThreadPriority(t, THREAD_PRIORITY_BELOW_NORMAL);
			break;
		case MYTHREAD_PRIORITY_NORMAL:
			SetThreadPriority(t, THREAD_PRIORITY_NORMAL);
			break;
		case MYTHREAD_PRIORITY_ABOVE_NORMAL:
			SetThreadPriority(t, THREAD_PRIORITY_ABOVE_NORMAL);
			break;
	};
}

#endif

// Wait for all the threads in a vector to finish
void join_all_threads(VECTOR_THREAD* pVectorThread)
{
	if (pVectorThread == NULL)
		return;

	for (VECTOR_THREAD_ITER iter = pVectorThread->begin();
		 iter != pVectorThread->end();
		 iter++)
	{
		join_thread(*iter);
	}
}


// Delete all the threads in a vector to finish
void delete_all_threads(VECTOR_THREAD* pVectorThread)
{
	if (pVectorThread == NULL)
		return;

	for (VECTOR_THREAD_ITER iter = pVectorThread->begin();
		 iter != pVectorThread->end();
		 iter++)
	{
		delete_thread(*iter);
	}
}

///////////////////////////// MutexBlock class implemetation /////////////////////////////

// Wait until we get the lock
MutexBlock::MutexBlock(mutex_t* pMutex)
{
	m_bIsLocked		= false;
	m_pMutex		= NULL;

	mutex_lock(pMutex);
	m_bIsLocked = true;
	m_pMutex = pMutex;
}
MutexBlock::~MutexBlock()
{
	if (m_bIsLocked)
		mutex_unlock(m_pMutex);
	m_bIsLocked = false;
}

MutexBlock::MutexBlock(mutex_t* pMutex, int iWaitMs)
{
	m_bIsLocked		= false;
	m_pMutex		= NULL;

	if (mutex_lock_timed(pMutex, iWaitMs))
	{
		m_bIsLocked = true;
		m_pMutex	= pMutex;
	}
}

// Find out if the mutex was obtained (only needs to be checked when
// a timeout was specified to the constructor.
bool MutexBlock::GetIsLocked()
{
	return m_bIsLocked;
}
