// Threading
// Cross-platform threading and mutual exclusion code.

#ifndef _THEADING_H_
#define _THEADING_H_

#include <vector>
#include <iostream>
#include <string.h>

// Stuff needed for threading on both Windows and linux
#ifdef WIN32
#include <windows.h>
#define THREAD_START DWORD WINAPI
#define COND_TIMEDOUT WAIT_TIMEOUT
typedef HANDLE condition_t;
typedef HANDLE mythread_t;
typedef HANDLE mutex_t;
#define cond_wait(c) WaitForSingleObject(c, INFINITE)
#define cond_wait_timed(cc,ticks)  (WaitForSingleObject(*(cc),ticks) == WAIT_OBJECT_0)
#define cond_signal(c) SetEvent(c)
#define create_cond(cc) (*(cc) = CreateEvent(NULL, TRUE, FALSE, NULL))
#define create_thread(tt, proc) (*(tt) = CreateThread(NULL, 0, proc, NULL, 0, NULL))
#define create_thread_stack(tt, proc, size) (*(tt) = CreateThread(NULL, size, proc, NULL, 0, NULL))
#define create_thread_stack_arg(tt, proc, size, arg) (*(tt) = CreateThread(NULL, size, proc, (void*) arg, 0, NULL))
#define create_thread_arg(tt, proc, arg) (*(tt) = CreateThread(NULL, 0, proc, (void *) arg, 0, NULL))
#define join_thread(t) WaitForSingleObject(t, INFINITE)
#define join_thread_timed(t,ticks) (WaitForSingleObject(t, ticks) == WAIT_OBJECT_0)
#define delete_thread(t) CloseHandle(t)
#define kill_thread(t) TerminateThread(t, 0)
#define sleep_ms(ms) Sleep(ms);
typedef DWORD thread_id;
#define get_current_thread_id() GetCurrentThreadId();

// Added to provide mutual exclusion code blocks
typedef HANDLE mutex_t;
#define create_mutex(m) *m = CreateMutex(NULL, FALSE, NULL);
#define mutex_lock(m) WaitForSingleObject(*m, INFINITE);
#define mutex_lock_timed(m, ms) (WaitForSingleObject(*m, ms) == WAIT_OBJECT_0)
#define mutex_unlock(m) ReleaseMutex(*m);
#define mutex_delete(m) CloseHandle(*m);

#else                           /* !WIN32 */

#include <pthread.h>
#include <sys/time.h>
#include <time.h>
#include <sched.h>
#include <stdio.h>

#define THREAD_START void *
#define COND_TIMEDOUT ETIMEDOUT
typedef struct {
    pthread_cond_t cond;
    pthread_mutex_t mtx;
} condition_t;
typedef pthread_t mythread_t;
#define cond_wait(c) {				\
	pthread_mutex_lock(&(c).mtx);		\
	pthread_cond_wait(&(c).cond, &(c).mtx);	\
	pthread_mutex_unlock(&(c).mtx);		\
}

bool cond_wait_timed(condition_t * c, int ticks);
void sleep_ms(long msec);

#define cond_signal(c) {			\
	pthread_mutex_lock(&(c).mtx);		\
	pthread_cond_signal(&(c).cond);		\
	pthread_mutex_unlock(&(c).mtx);		\
}
#define create_cond(cc) {			\
	pthread_cond_init(&(cc)->cond, NULL);	\
	pthread_mutex_init(&(cc)->mtx, NULL);	\
}
#define create_thread(tt, proc) pthread_create(tt, NULL, proc, NULL)
#define create_thread_arg(tt, proc, arg) pthread_create(tt, NULL, proc, arg)
void create_thread_stack(pthread_t* tt, void *(proc)(void*), int size);
void create_thread_stack_arg(pthread_t* tt, void *(proc)(void*), int size, void* arg);
#define join_thread(t) pthread_join(t, NULL)
#define delete_thread(t)
#define kill_thread(t) pthread_kill(t, 0)

// Added to provide mutual exclusion code blocks.
// These all take a pointer to the mutex object.
typedef pthread_mutex_t mutex_t;
#define create_mutex(m) pthread_mutex_init(m, NULL);
#define mutex_lock(m) 	pthread_mutex_lock(m);
#define mutex_unlock(m) pthread_mutex_unlock(m);
#define mutex_delete(m) pthread_mutex_destroy(m);

typedef pthread_t thread_id;
#define get_current_thread_id() pthread_self();
bool mutex_lock_timed(mutex_t* pMutex, int iTimeoutMs);

#endif                          /* !WIN32 */

// Types for dealing with vectors of threads
typedef std::vector<mythread_t> VECTOR_THREAD;
typedef std::vector<mythread_t>::iterator VECTOR_THREAD_ITER;

void join_all_threads(VECTOR_THREAD* pVectorThread);
void delete_all_threads(VECTOR_THREAD* pVectorThread);

// Simple class that controls mutual exclusion by scoping of an object.
// It can either block undefintely, or wait a specified amount (in this
// case the caller needs to check its state to see if a lock was 
// obtained);
class MutexBlock
{
	public:
		// Wait until we get the lock
		MutexBlock(mutex_t* pMutex);
		MutexBlock(mutex_t* pMutex, int iWaitMs);
		~MutexBlock();

		bool GetIsLocked();

	private:
		mutex_t*		m_pMutex;
		bool			m_bIsLocked;
};

// Stuff for setting a threads priority
#define MYTHREAD_PRIORITY_BELOW_NORMAL 0
#define MYTHREAD_PRIORITY_NORMAL 1
#define MYTHREAD_PRIORITY_ABOVE_NORMAL 2
void set_thread_priority(mythread_t t, int iPriority);


#endif

